@@ -1,16 +1,22 @@
 import logging
 from decimal import Decimal, ROUND_DOWN, getcontext
-from typing import Union, List, Dict
+from typing import Union, List, Dict, Optional
 from privex.jsonrpc import SteemEngineRPC
 from privex.steemengine import exceptions
 from privex.steemengine.SteemEngineHistory import SteemEngineHistory
-from privex.steemengine.objects import SEBalance, SETransaction, Token, SETrade, SEOrder
-from privex.helpers import empty
+from privex.steemengine.exceptions import NoResults
+from privex.steemengine.objects import SEBalance, SETransaction, Token, SETrade, SEOrder, SETransactionInfo, conv_dec, AnyNum, SEPlacedOrder
+from privex.helpers import empty, dec_round, r_cache, DictObject
 
 log = logging.getLogger(__name__)
 getcontext().rounding = ROUND_DOWN
 
 
+def round_str(amount: AnyNum, dp: AnyNum = 2) -> str:
+    """Round ``amount`` to ``dp`` decimal places as a string"""
+    return ('{0:.' + str(int(dp)) + 'f}').format(conv_dec(amount))
+
+
 class SteemEngineToken:
     """
     SteemEngineToken - a wrapper class around :py:class:`privex.jsonrpc.SteemEngineRPC`, with support
@@ -50,7 +56,9 @@ class SteemEngineToken:
     }
 
     network: str
-
+    native_coin: str
+    DEFAULT_BLOCKCHAIN_URL = '/rpc/blockchain'
+    
     @property
     def steem(self):
         """
@@ -86,18 +94,33 @@ class SteemEngineToken:
         """
         rpc_args = dict(rpc_args)
         history_conf = {} if history_conf is None else history_conf
+        network = network.lower()
         self.network = network
+        native_coin = 'STEEMP'
         if network == "hive":
             rpc_args['hostname'] = rpc_args.get('hostname', 'api.hive-engine.com')
             rpc_args['url'] = rpc_args.get('url', '/rpc/contracts')
+            network_account = 'ssc-mainnet-hive'
             history_conf['hostname'] = history_conf.get('hostname', 'accounts.hive-engine.com')
             history_conf['url'] = history_conf.get('url', 'accountHistory')
-
-        self.rpc = SteemEngineRPC(**rpc_args)
+            native_coin = 'SWAP.HIVE'
+        
+        self.native_coin = rpc_args.pop('native_coin', native_coin).upper()
+        
+        self.rpc = SteemEngineRPC(**dict(rpc_args))
+        if 'url' in rpc_args:
+            del rpc_args['url']
+        self.block_rpc = SteemEngineRPC(**rpc_args, url=self.DEFAULT_BLOCKCHAIN_URL)
         self.network_account = network_account
         self.history_rpc = SteemEngineHistory(**history_conf)
         log.debug('Initialized SteemEngineToken with args: %s %s %s', network_account, history_conf, rpc_args)
 
+    @property
+    @r_cache(lambda self: f"pvx_seng:token:{self.native_coin}")
+    def native_token(self) -> Token:
+        """Returns the :class:`.Token` object for the native coin :attr:`.native_coin`"""
+        return self.get_token(self.native_coin)
+
     @staticmethod
     def custom_beem(node: Union[str, list] = "", *args, **kwargs):
         """
@@ -154,12 +177,13 @@ class SteemEngineToken:
         """
         balances = self.get_balances(user)
         for bal in balances:
-            if bal['symbol'] == symbol.upper():
-                log.debug('Found token balance matching %s, returning %s', symbol, bal['balance'])
-                return Decimal(bal['balance'])
+            if bal.symbol == symbol.upper():
+                log.debug('Found token balance matching %s, returning %s', symbol, bal.balance)
+                return bal.balance
         log.debug('Did not find token balance matching %s, returning 0', symbol)
         return Decimal(0)
 
+    @r_cache(lambda self, user: f"pvx_seng:account_exists:{user}", 30)
     def account_exists(self, user) -> bool:
         """
         Helper function to verify if a given user exists on Steem.
@@ -178,6 +202,7 @@ class SteemEngineToken:
         log.debug('Checking if user %s exists', user)
         return len(self.steem.rpc.get_account(user)) > 0
 
+    @r_cache(lambda self, limit=1000, offset=0: f"pvx_seng:list_tokens:{limit}:{offset}", 120)
     def list_tokens(self, limit=1000, offset=0) -> List[Token]:
         """
         Returns a list of all tokens as :class:`.Token` objects.
@@ -216,74 +241,156 @@ class SteemEngineToken:
             limit=limit, offset=offset
         )))
 
-    def order_history(self, symbol, limit=30, offset=0, indexes: List[dict] = None) -> List[SETrade]:
+    def query_order_history(self, limit=30, offset=0, indexes: List[dict] = None, **query) -> List[SETrade]:
+        """
+        Used internally by methods such as :meth:`.order_history` and :meth:`.find_fulfilled_sells` etc.
+        
+        Queries the ``market`` contract table ``tradeHistory`` using the query params passed as kwargs via ``query``.
+        
+        :param int limit: Amount of orders to retrieve
+        :param int offset: Offset selection by this many rows (for pagination)
+        :param list indexes: A list of dictionary indexes, e.g. ``[dict(descending=False, index='timestamp')]``
+        :param query: Query parameters to filter order history
+        :return List[SETrade] orders: A list of :py:class:`.SETrade` objects
+        """
+        indexes = [dict(descending=False, index='_id')] if empty(indexes) else indexes
+        data = dict(
+            contract='market', table='tradesHistory',
+            query=dict(query), indexes=indexes,
+            limit=limit, offset=offset
+        )
+        res = self.rpc.find(**data)
+        if empty(res):
+            raise NoResults("query_order_history got empty (None) response from API...")
+        return list(SETrade.from_list(res, seng_ins=self))
+    
+    def order_history(self, symbol, limit=30, offset=0, indexes: List[dict] = None, **query) -> List[SETrade]:
         """
         Get a list of recent Steem Engine orders for a given symbol.
 
         **Example:**
 
+            >>> st = SteemEngineToken()
             >>> o = SteemEngineToken().order_history('ENG')
             >>> o[0]
             <SETrade type='buy' price='0.99' symbol='ENG' quantity='0.80405854'>
             >>> o[0].timestamp
             '2019-07-27 01:06:09'
+            
+            >>> st = SteemEngineToken(network='hive')
+            >>> ht = st.order_history('BEE', sellTxId='d442b0156ccddb0d6fca381fa0a8ea027b318d17')
+            >>> dict(ht[0])
+            {'_id': 61528, 'type': 'sell',
+             'buyer': 'arquitectojm', 'seller': 'cryptomancer', 'symbol': 'BEE',
+             'quantity': Decimal('3'), 'price': Decimal('0.94900000'),
+             'timestamp': datetime.datetime(2020, 6, 21, 8, 0, 9, tzinfo=tzutc()),
+             'volume': Decimal('2.84700000'),
+             'buyTxId': '6e0d09cd45e94eaf4fb2b85cdc0b9d3a28401cbd',
+             'sellTxId': 'd442b0156ccddb0d6fca381fa0a8ea027b318d17',
+             'direction': 'sell'
+             }
+
 
         :param str symbol: The symbol to get historic orders for
         :param int limit: Amount of orders to retrieve
         :param int offset: Offset selection by this many rows (for pagination)
+        :param query: Additional query parameters to filter order history
         :param list indexes: A list of dictionary indexes, e.g. ``[dict(descending=False, index='timestamp')]``
         :return List[SETrade] orders: A list of :py:class:`.SETrade` objects
         """
-        indexes = [dict(descending=False, index='timestamp')] if empty(indexes) else indexes
-        return list(SETrade.from_list(self.rpc.find(
-            contract='market',
-            table='tradesHistory',
-            query=dict(symbol=symbol),
-            indexes=indexes,
-            limit=limit, offset=offset
-        )))
+        return self.query_order_history(
+            limit=limit, offset=offset, indexes=indexes, symbol=symbol, **query
+        )
+
+    trade_history = order_history
+
+    def find_fulfilled_sells(self, txid: str, limit=30, offset=0, indexes: List[dict] = None, **query) -> List[SETrade]:
+        return self.query_order_history(
+            limit=limit, offset=offset, indexes=indexes, sellTxId=txid, **query
+        )
+
+    def find_fulfilled_buys(self, txid: str, limit=30, offset=0, indexes: List[dict] = None, **query) -> List[SETrade]:
+        return self.query_order_history(
+            limit=limit, offset=offset, indexes=indexes, buyTxId=txid, **query
+        )
+
+    def find_fulfilled(self, txid: str, limit=30, offset=0, indexes: List[dict] = None, **query) -> List[SETrade]:
+        buys = self.find_fulfilled_buys(txid=txid, limit=limit, offset=offset, indexes=indexes, **query)
+        sells = self.find_fulfilled_sells(txid=txid, limit=limit, offset=offset, indexes=indexes, **query)
+        return buys + sells
     
     def get_orderbook(self, symbol, direction='buy', user=None, limit=200, offset=0,
-                      indexes: list = None) -> List[SEOrder]:
+                      indexes: list = None, sort_by: str = 'price', sort_reverse=None, **query) -> List[SEOrder]:
         """
-        Get a list of open Steem Engine orders for a given symbol, by default will display ``'buy'`` orders unless
+        Get a list of open Steem/Hive Engine orders for a given symbol, by default will display ``'buy'`` orders unless
         you set ``direction`` to ``'sell'``
-
-        **Example:**
-
-            >>> o = SteemEngineToken().get_orderbook('ENG', direction='sell')
+        
+        NOTE: By default, if ``direction`` is ``'buy'``, then the orders will be returned sorted by price - descending (highest first),
+              and for ``sell`` - ascending (lowest first).
+        
+        **Example**::
+            
+            >>> st = SteemEngineToken()
+            >>> o = st.get_orderbook('ENG', direction='sell')
             >>> o[0]
             <SEOrder account='someguy123' price='0.99' symbol='ENG' quantity='885.40249121'>
             >>> str(o[0].timestamp)
             '2019-07-26 10:46:18'
+        
+        **Sort by highest quantity first**::
+            
+            >>> st.get_orderbook('BEE', 'sell', sort_by='quantity', sort_reverse=True)
+        
+        **Sort by oldest order first**::
+            
+            >>> st.get_orderbook('BEE', 'sell', sort_by='timestamp')
 
-
-        :param str symbol: The symbol to get the open orders for
-        :param int limit: Amount of orders to retrieve
-        :param int offset: Offset selection by this many rows (for pagination)
-        :param list indexes: A list of dictionary indexes, e.g. ``[dict(descending=False, index='timestamp')]``
-        :param str user: If ``None`` , get all orders, otherwise only get orders by this user. Default: ``None``
-        :param str direction: The direction of orders to get, either ``buy`` or ``sell`` Default: ``buy``
+        :param bool sort_reverse:   ``True`` to reverse sort ``sort_by`` (descending), ``False`` to NOT reverse (ascending). The default
+                                    ``None`` will sort descending (reversed) when ``direction`` is ``buy``, and ascending for ``sell``
+                                    (direction-based sorting only applies when ``sort_by`` is ``'price'``)
+        
+        :param str sort_by:         (Default: ``price``) The attribute of :class:`.SEOrder` to sort by.
+        :param str symbol:          The symbol to get the open orders for
+        :param int limit:           Amount of orders to retrieve
+        :param int offset:          Offset selection by this many rows (for pagination)
+        :param list indexes:        A list of dictionary indexes, e.g. ``[dict(descending=False, index='timestamp')]``
+        :param str user:            If ``None`` , get all orders, otherwise only get orders by this user. Default: ``None``
+        :param str direction:       The direction of orders to get, either ``buy`` or ``sell`` Default: ``buy``
+        
+        :raises NoResult:           Raised when ``None`` is returned by ``rpc.find``
+        
         :return List[SEOrder] orders: A list of :py:class:`.SEOrder` objects
         """
         direction = direction.lower()
         if direction not in ['buy', 'sell']:
             raise AttributeError('get_orderbook direction must be either "buy" or "sell"')
 
-        indexes = [dict(descending=direction == 'sell', index='price')] if empty(indexes) else indexes
+        # indexes = [dict(descending=direction == 'sell', index='price')] if empty(indexes) else indexes
+        indexes = [] if empty(indexes) else indexes
 
-        q = dict(symbol=symbol)
+        q = dict(symbol=symbol, **query)
         if not empty(user):
             q['account'] = user
-        return list(SEOrder.from_list(self.rpc.find(
+        res = self.rpc.find(
             contract='market',
             table=f'{direction.lower()}Book',
             query=q,
             indexes=indexes,
             limit=limit, offset=offset
-        )))
+        )
+        if empty(res):
+            raise NoResults("get_orderbook got empty (None) response from API...")
+        orders = list(SEOrder.from_list(res, seng_ins=self))
+        
+        if len(indexes) == 0 and sort_by == 'price' and sort_reverse is None:
+            sort_reverse = False if direction == 'sell' else True
+        sort_reverse = False if sort_reverse is None else sort_reverse
+        if not empty(sort_by):
+            orders = sorted(orders, key=lambda p: getattr(p, sort_by), reverse=sort_reverse)
+        
+        return list(orders)
 
-    def find_steem_tx(self, tx_data: dict, last_blocks=15) -> dict:
+    def find_steem_tx(self, tx_data: dict, last_blocks=15) -> Optional[dict]:
         """
         Used internally to get the transaction ID after a Steem transaction has been broadcasted.
 
@@ -339,7 +446,8 @@ class SteemEngineToken:
         log.debug('Getting TX history for user %s, symbol %s, limit %s, offset %s', user, symbol, limit, offset)
         return list(SETransaction.from_list(self.history_rpc.get_history(account=user, symbol=symbol, limit=limit, offset=offset)))
 
-    def get_token(self, symbol) -> Token:
+    @r_cache(lambda self, symbol: f"pvx_seng:token:{symbol}", 60)
+    def get_token(self, symbol) -> Optional[Token]:
         """
         Get the token object for an individual token.
 
@@ -378,7 +486,7 @@ class SteemEngineToken:
             table='tokens',
             query=dict(symbol=symbol.upper())
         )
-        return None if empty(tk) else Token(**tk)
+        return None if empty(tk) else Token.from_dict(tk)
 
     def send_token(self, symbol, from_acc, to_acc, amount: Decimal, memo="", find_tx=True) -> dict:
         """
@@ -547,12 +655,111 @@ class SteemEngineToken:
             tx = self.find_steem_tx(j)
             return j if tx is None else tx
         return j
+    
+    def place_order(self, user: str, action: str, symbol: str, quantity: AnyNum, price: AnyNum, find_tx=True) -> SEPlacedOrder:
+        """
+        Place an order on the Steem/Hive Engine market
+        
+        **Placing a 0.01 BEE sell order for 0.9 SWAP.HIVE per BEE from the account someguy123**::
+        
+            >>> from privex.steemengine import SteemEngineToken
+            >>> s = SteemEngineToken(network='hive')
+            >>> order = s.place_order('someguy123', 'sell', 'BEE', '0.01', '0.9')
+            >>> order.user
+            'someguy123'
+            >>> order.events
+            [
+                SETransactionLogEvent(contract='tokens', event='transferToContract',
+                    data=SEContractTransfer(sender='someguy123', to='market', symbol='BEE', quantity=Decimal('0.01000000'))),
+                SETransactionLogEvent(contract='tokens', event='transferFromContract',
+                    data=SEContractTransfer(sender='market', to='everything-4you', symbol='BEE', quantity=Decimal('0.01000000'))),
+                SETransactionLogEvent(contract='tokens', event='transferFromContract',
+                    data=SEContractTransfer(sender='market', to='someguy123', symbol='SWAP.HIVE', quantity=Decimal('0.00902100'))),
+                SETransactionLogEvent(contract='market', event='orderClosed',
+                    data={'account': 'someguy123', 'type': 'sell', 'txId': '2c5993eacf37c16e9a34ba7a35a0a5efa7fbc729'})
+            ]
+        
+        
+        :param str user: The user to place the order with.
+        :param str action: Either ``'buy'`` or ``'sell'``
+        :param str symbol: The symbol of the token you want to ``'buy'`` or ``'sell'``
+        :param Decimal|int|float|str quantity: The amount of ``symbol`` to buy/sell
+        :param Decimal|int|float|str price: The price in STEEMP/SWAP.HIVE per 1 ``symbol``
+        :param bool find_tx: (Default: ``True``) Verify the transaction was successful by searching for it on the blockchain
+        :return SEPlacedOrder result: The details of the order which was placed
+        """
+        # {"contractName": "market", "contractAction": "sell", "contractPayload": {"symbol": "BEE", "quantity": "1", "price": "0.7"}}
+        action, symbol = action.lower(), symbol.upper()
+        if action not in ['buy', 'sell']:
+            raise AttributeError("place_order expects action to be either 'buy' or 'sell'")
+        
+        t = self.get_token(symbol)
+        quantity, price = conv_dec(quantity), conv_dec(price)
+        
+        if t is None:
+            raise exceptions.TokenNotFound('Token {} does not exist'.format(t))
+        prec = t.precision
+        if quantity < Decimal(pow(10, -prec)):
+            log.warning('Quantity %s was passed, but is lower than precision for %s', quantity, symbol)
+            raise ArithmeticError(f'Quantity {quantity} is lower than token {symbol}s precision of {t["precision"]} DP')
+        if price < Decimal(pow(10, -prec)):
+            log.warning('Price %s was passed, but is lower than precision for %s', price, symbol)
+            raise ArithmeticError(f'Price {price} is lower than token {symbol}s precision of {t["precision"]} DP')
+        
+        quantity, price = dec_round(quantity, prec), dec_round(price, prec)
+        if action == 'sell':
+            # For sell orders, the user is using 'symbol' for the order, so we check 'symbol' balance
+            bal = self.get_token_balance(user=user, symbol=symbol)
+            if bal < quantity:
+                raise exceptions.NotEnoughBalance(
+                    f"User {user} tried to place an sell order for {quantity} {symbol} but only has {bal} {symbol}"
+                )
+        else:
+            # For buy orders, the user is using the native coin of the network (e.g. STEEMP or SWAP.HIVE),
+            # so we have to check their balance of self.native_coin, and sum the amount of native_coin that would be needed.
+            bal = self.get_token_balance(user=user, symbol=self.native_coin)
+            total_native = dec_round(quantity * price, self.native_token.precision)
+            if bal < total_native:
+                raise exceptions.NotEnoughBalance(
+                    f"User {user} tried to place an buy order for {quantity} {symbol} ({total_native} {self.native_coin}) "
+                    f"but only has {bal} {self.native_coin}"
+                )
+            
+        custom = DictObject(
+            contractName="market",
+            contractAction=action,
+            contractPayload=DictObject(
+                symbol=symbol,
+                quantity=round_str(quantity, prec),
+                price=round_str(price, self.native_token.precision)
+            )
+        )
+        ctpayload = custom.contractPayload
+        res = SEPlacedOrder(
+            symbol=symbol, quantity=ctpayload.quantity, price=ctpayload.price, user=user,
+            direction=action, custom_tx=custom, _seng_instance=self
+        )
+        log.debug("Broadcasting custom_json from %s to %s - data: %s", user, self.network_account, custom)
+        j = self.steem.custom_json(self.network_account, custom, [user])
+        res.broadcast_result = j
+        
+        if find_tx:  # Find the transaction on the blockchain to get TXID and other data from it.
+            log.debug("Finding transaction: %s", j)
+            tx = self.find_steem_tx(j)
+            j = j if tx is None else tx
+            res.network_transaction = tx
+            res.txid = j.get('transaction_id')
+            
+        return res
 
+    @r_cache(lambda self, txid: f"pvx_seng:tx_info:{txid}", 30)
+    def get_transaction_info(self, txid: str) -> SETransactionInfo:
+        return SETransactionInfo.from_dict(self.block_rpc.getTransactionInfo(txid=txid))
 
 
 """
 +===================================================+
-|                 © 2019 Privex Inc.                |
+|                 © 2020 Privex Inc.                |
 |               https://www.privex.io               |
 +===================================================+
 |                                                   |
@@ -566,7 +773,7 @@ class SteemEngineToken:
 +===================================================+
 
 Python SteemEngine - A small library for querying and interacting with the SteemEngine network (https://steem-engine.com)
-Copyright (c) 2019    Privex Inc. ( https://www.privex.io )
+Copyright (c) 2020    Privex Inc. ( https://www.privex.io )
 
 Permission is hereby granted, free of charge, to any person obtaining a copy of this software and associated documentation 
 files (the "Software"), to deal in the Software without restriction, including without limitation the rights to use, copy, 
